"""Variable definitions"""

# pylint: disable=too-many-lines
from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from typing import Iterable, Literal, NamedTuple

import netCDF4


class Product(Enum):
    # Level 1b
    RADAR = "radar"
    LIDAR = "lidar"
    MWR = "mwr"
    DISDROMETER = "disdrometer"
    MODEL = "model"
    WEATHER_STATION = "weather-station"
    DOPPLER_LIDAR = "doppler-lidar"
    DOPPLER_LIDAR_WIND = "doppler-lidar-wind"
    RAIN_RADAR = "rain-radar"
    # Level 1c
    CATEGORIZE = "categorize"
    MWR_L1C = "mwr-l1c"
    CATEGORIZE_VOODOO = "categorize-voodoo"
    # Level 2
    CLASSIFICATION = "classification"
    CLASSIFICATION_VOODOO = "classification-voodoo"
    IWC = "iwc"
    LWC = "lwc"
    DER = "der"
    IER = "ier"
    DRIZZLE = "drizzle"
    MWR_SINGLE = "mwr-single"
    MWR_MULTI = "mwr-multi"
    # Level 3
    L3_CF = "l3-cf"
    L3_IWC = "l3-iwc"
    L3_LWC = "l3-lwc"

    @classmethod
    def all(cls) -> set[Product]:
        return set(cls)


Level = Literal["1b", "1c", "2", "3"]

LEVELS: dict[Product, Level] = {
    Product.RADAR: "1b",
    Product.LIDAR: "1b",
    Product.MWR: "1b",
    Product.DISDROMETER: "1b",
    Product.MODEL: "1b",
    Product.WEATHER_STATION: "1b",
    Product.DOPPLER_LIDAR: "1b",
    Product.DOPPLER_LIDAR_WIND: "1c",
    Product.RAIN_RADAR: "1b",
    Product.CATEGORIZE: "1c",
    Product.CATEGORIZE_VOODOO: "1c",
    Product.MWR_L1C: "1c",
    Product.CLASSIFICATION: "2",
    Product.CLASSIFICATION_VOODOO: "2",
    Product.IWC: "2",
    Product.LWC: "2",
    Product.DER: "2",
    Product.IER: "2",
    Product.DRIZZLE: "2",
    Product.MWR_SINGLE: "2",
    Product.MWR_MULTI: "2",
    Product.L3_CF: "3",
    Product.L3_IWC: "3",
    Product.L3_LWC: "3",
}


class Dtype(str, Enum):
    FLOAT = "float32"
    DOUBLE = "float64"
    INT = "int32"
    SHORT = "int16"
    BYTE = "int8"


class Variable(NamedTuple):
    long_name: str
    units: str | Callable[[netCDF4.Dataset], str] | None = "1"
    dtype: str = Dtype.FLOAT
    standard_name: str | None = None
    required: Iterable[Product] | None = None


def time_units(nc: netCDF4.Dataset) -> str:
    return f"hours since {nc.year}-{nc.month}-{nc.day} 00:00:00 +00:00"


VARIABLES = {
    # -------------------------------
    # Required in RADAR Level 1b file
    # -------------------------------
    "radar_frequency": Variable(
        long_name="Radar transmit frequency",
        units="GHz",
        required=[Product.RADAR],
    ),
    "Zh": Variable(
        long_name="Radar reflectivity factor",
        units="dBZ",
        required=[
            Product.RADAR,
            Product.RAIN_RADAR,
        ],
    ),
    "nyquist_velocity": Variable(
        long_name="Nyquist velocity",
        units="m s-1",
        required=[Product.RADAR],
    ),
    # -------------------------------
    # Required in LIDAR Level 1b file
    # -------------------------------
    "wavelength": Variable(
        long_name="Laser wavelength",
        units="nm",
        required=[Product.LIDAR],
    ),
    "zenith_angle": Variable(
        long_name="Zenith angle",
        units="degree",
        standard_name="zenith_angle",
        required=[Product.LIDAR],
    ),
    # -------------------------------------
    # Required in DISDROMETER Level 1b file
    # -------------------------------------
    "radar_reflectivity": Variable(
        long_name="Equivalent radar reflectivity factor",
        units="dBZ",
        standard_name="equivalent_reflectivity_factor",
        required=[Product.DISDROMETER],
    ),
    "visibility": Variable(
        long_name="Meteorological optical range (MOR) visibility",
        units="m",
        standard_name="visibility_in_air",
        dtype=Dtype.INT,
        required=[Product.DISDROMETER],
    ),
    "n_particles": Variable(
        long_name="Number of particles in time interval",
        dtype=Dtype.INT,
    ),
    # -----------------------------------------
    # Required in WEATHER STATION Level 1b file
    # -----------------------------------------
    "air_temperature": Variable(
        long_name="Air temperature",
        standard_name="air_temperature",
        units="K",
        required=[Product.WEATHER_STATION],
    ),
    "air_pressure": Variable(
        long_name="Air pressure",
        units="Pa",
        standard_name="air_pressure",
        required=[Product.WEATHER_STATION],
    ),
    "rainfall_amount": Variable(
        long_name="Rainfall amount",
        units="m",
        standard_name="thickness_of_rainfall_amount",
        required=[Product.WEATHER_STATION],
    ),
    # ------------------------------------
    # Required in RAIN RADAR Level 1b file
    # ------------------------------------
    "pia": Variable(
        long_name="Path integrated rain attenuation",
        units="dB",
        required=[Product.RAIN_RADAR],
    ),
    # ------------------------------------
    # Required in CATEGORIZE Level 1c file
    # ------------------------------------
    "lidar_wavelength": Variable(
        long_name="Laser wavelength",
        units="nm",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "insect_prob": Variable(
        long_name="Insect probability",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "uwind": Variable(
        long_name="Zonal wind",
        units="m s-1",
        required=[
            Product.CATEGORIZE,
            Product.CATEGORIZE_VOODOO,
            Product.DOPPLER_LIDAR_WIND,
        ],
    ),
    "uwind_raw": Variable(
        long_name="Non-screened zonal wind",
        units="m s-1",
    ),
    "vwind": Variable(
        long_name="Meridional wind",
        units="m s-1",
        required=[
            Product.CATEGORIZE,
            Product.CATEGORIZE_VOODOO,
            Product.DOPPLER_LIDAR_WIND,
        ],
    ),
    "vwind_raw": Variable(
        long_name="Non-screened meridional wind",
        units="m s-1",
    ),
    "azimuth_offset": Variable(
        long_name="Azimuth offset of the instrument (positive clockwise from north)",
        units="degrees",
    ),
    "q": Variable(
        long_name="Specific humidity",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "Tw": Variable(
        long_name="Wet-bulb temperature",
        units="K",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "category_bits": Variable(
        long_name="Target categorization bits",
        dtype=Dtype.INT,
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "radar_liquid_atten": Variable(
        long_name="Two-way radar attenuation due to liquid water",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "radar_gas_atten": Variable(
        long_name="Two-way radar attenuation due to atmospheric gases",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "radar_rain_atten": Variable(
        long_name="Two-way radar attenuation due to rain",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "radar_melting_atten": Variable(
        long_name="Two-way radar attenuation due to melting ice",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "quality_bits": Variable(
        long_name="Data quality bits",
        dtype=Dtype.INT,
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "beta": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
        # Lidar has its own beta test
        required=[
            Product.CATEGORIZE,
            Product.CATEGORIZE_VOODOO,
            Product.DOPPLER_LIDAR,
        ],
    ),
    "beta_error": Variable(
        long_name="Error in attenuated backscatter coefficient",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "beta_bias": Variable(
        long_name="Bias in attenuated backscatter coefficient",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "v_sigma": Variable(
        long_name="Standard deviation of mean Doppler velocity",
        units="m s-1",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "Z": Variable(
        long_name="Radar reflectivity factor",
        units="dBZ",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "Z_bias": Variable(
        long_name="Bias in radar reflectivity factor",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "Z_error": Variable(
        long_name="Error in radar reflectivity factor",
        units="dB",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "Z_sensitivity": Variable(
        long_name="Minimum detectable radar reflectivity",
        units="dBZ",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "model_time": Variable(
        long_name="Model time UTC",
        units=time_units,
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    "model_height": Variable(
        long_name="Height of model variables above mean sea level",
        units="m",
        required=[Product.CATEGORIZE, Product.CATEGORIZE_VOODOO],
    ),
    # ---------------------------------------
    # Required in CLASSIFICATION Level 2 file
    # ---------------------------------------
    "target_classification": Variable(
        long_name="Target classification",
        required=[Product.CLASSIFICATION, Product.CLASSIFICATION_VOODOO],
        dtype=Dtype.INT,
    ),
    "detection_status": Variable(
        long_name="Radar and lidar detection status",
        required=[Product.CLASSIFICATION, Product.CLASSIFICATION_VOODOO],
        dtype=Dtype.INT,
    ),
    "cloud_base_height_amsl": Variable(
        long_name="Height of cloud base above mean sea level",
        units="m",
        required=[Product.CLASSIFICATION, Product.CLASSIFICATION_VOODOO],
    ),
    "cloud_top_height_amsl": Variable(
        long_name="Height of cloud top above mean sea level",
        units="m",
        required=[Product.CLASSIFICATION, Product.CLASSIFICATION_VOODOO],
    ),
    "cloud_base_height_agl": Variable(
        long_name="Height of cloud base above ground level",
        units="m",
        required=[Product.CLASSIFICATION, Product.CLASSIFICATION_VOODOO],
    ),
    "cloud_top_height_agl": Variable(
        long_name="Height of cloud top above ground level",
        units="m",
        required=[Product.CLASSIFICATION, Product.CLASSIFICATION_VOODOO],
    ),
    # ----------------------------
    # Required in LWC Level 2 file
    # ----------------------------
    "lwc": Variable(
        long_name="Liquid water content",
        units="kg m-3",
        required=[
            Product.LWC,
            Product.RAIN_RADAR,
        ],
    ),
    "lwc_error": Variable(
        long_name="Random error in liquid water content, one standard deviation",
        units="dB",
        required=[Product.LWC],
    ),
    "lwc_retrieval_status": Variable(
        long_name="Liquid water content retrieval status",
        dtype=Dtype.INT,
        required=[Product.LWC],
    ),
    # ----------------------------
    # Required in IWC Level 2 file
    # ----------------------------
    "iwc": Variable(
        long_name="Ice water content",
        units="kg m-3",
        required=[Product.IWC],
    ),
    "iwc_error": Variable(
        long_name="Random error in ice water content",
        units="dB",
        required=[Product.IWC],
    ),
    "iwc_bias": Variable(
        long_name="Possible bias in ice water content",
        units="dB",
        required=[Product.IWC],
    ),
    "iwc_sensitivity": Variable(
        long_name="Minimum detectable ice water content",
        units="kg m-3",
        required=[Product.IWC],
    ),
    "iwc_retrieval_status": Variable(
        long_name="Ice water content retrieval status",
        dtype=Dtype.INT,
        required=[Product.IWC],
    ),
    # Deprecated:
    "iwc_inc_rain": Variable(
        long_name="Ice water content including rain",
        units="kg m-3",
    ),
    # --------------------------------
    # Required in DRIZZLE Level 2 file
    # --------------------------------
    "Do": Variable(
        long_name="Drizzle median diameter",
        units="m",
        required=[Product.DRIZZLE],
    ),
    "mu": Variable(
        long_name="Drizzle droplet size distribution shape parameter",
        required=[Product.DRIZZLE],
    ),
    "S": Variable(
        long_name="Lidar backscatter-to-extinction ratio",
        units="sr",
        required=[Product.DRIZZLE],
    ),
    "beta_corr": Variable(
        long_name="Lidar backscatter correction factor",
        required=[Product.DRIZZLE],
    ),
    "drizzle_N": Variable(
        long_name="Drizzle number concentration",
        units="m-3",
        required=[Product.DRIZZLE],
    ),
    "drizzle_lwc": Variable(
        long_name="Drizzle liquid water content",
        units="kg m-3",
        required=[Product.DRIZZLE],
    ),
    "drizzle_lwf": Variable(
        long_name="Drizzle liquid water flux",
        units="kg m-2 s-1",
        required=[Product.DRIZZLE],
    ),
    "v_drizzle": Variable(
        long_name="Drizzle droplet fall velocity",
        units="m s-1",
        required=[Product.DRIZZLE],
    ),
    "v_air": Variable(
        long_name="Vertical air velocity",
        units="m s-1",
        required=[Product.DRIZZLE],
    ),
    "Do_error": Variable(
        units="dB",
        long_name="Random error in drizzle median diameter",
        required=[Product.DRIZZLE],
    ),
    "drizzle_lwc_error": Variable(
        units="dB",
        long_name="Random error in drizzle liquid water content",
        required=[Product.DRIZZLE],
    ),
    "drizzle_lwf_error": Variable(
        units="dB",
        long_name="Random error in drizzle liquid water flux",
        required=[Product.DRIZZLE],
    ),
    "S_error": Variable(
        long_name="Random error in lidar backscatter-to-extinction ratio",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "Do_bias": Variable(
        long_name="Possible bias in drizzle median diameter",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "drizzle_lwc_bias": Variable(
        long_name="Possible bias in drizzle liquid water content",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "drizzle_lwf_bias": Variable(
        long_name="Possible bias in drizzle liquid water flux",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "drizzle_N_error": Variable(
        long_name="Random error in drizzle number concentration",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "v_drizzle_error": Variable(
        long_name="Random error in drizzle droplet fall velocity",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "mu_error": Variable(
        long_name="Random error in drizzle droplet size distribution shape parameter",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "drizzle_N_bias": Variable(
        long_name="Possible bias in drizzle number concentration",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "v_drizzle_bias": Variable(
        long_name="Possible bias in drizzle droplet fall velocity",
        units="dB",
        required=[Product.DRIZZLE],
    ),
    "drizzle_retrieval_status": Variable(
        long_name="Drizzle parameter retrieval status",
        required=[Product.DRIZZLE],
        dtype=Dtype.INT,
    ),
    # ----------------------------
    # Required in IER Level 2 file
    # ----------------------------
    "ier": Variable(
        long_name="Ice effective radius",
        units="m",
        required=[Product.IER],
    ),
    "ier_error": Variable(
        long_name="Random error in ice effective radius",
        units="m",
        required=[Product.IER],
    ),
    "ier_retrieval_status": Variable(
        long_name="Ice effective radius retrieval status",
        dtype=Dtype.INT,
        required=[Product.IER],
    ),
    # Deprecated:
    "ier_inc_rain": Variable(
        long_name="Ice effective radius including rain",
        units="m",
    ),
    # ----------------------------
    # Required in DER Level 2 file
    # ----------------------------
    "der": Variable(
        long_name="Droplet effective radius",
        units="m",
        required=[Product.DER],
    ),
    "der_error": Variable(
        long_name="Absolute error in droplet effective radius",
        units="m",
        required=[Product.DER],
    ),
    "der_scaled": Variable(
        long_name="Droplet effective radius (scaled to LWP)",
        units="m",
        required=[Product.DER],
    ),
    "der_scaled_error": Variable(
        long_name="Absolute error in droplet effective radius (scaled to LWP)",
        units="m",
        required=[Product.DER],
    ),
    "N_scaled": Variable(
        long_name="Cloud droplet number concentration",
        required=[Product.DER],
    ),
    "der_retrieval_status": Variable(
        long_name="Droplet effective radius retrieval status",
        dtype=Dtype.INT,
        required=[Product.DER],
    ),
    # ------------------------
    # Required in MWR L1C file
    # ------------------------
    "frequency": Variable(
        long_name="Nominal centre frequency of microwave channels",
        units="GHz",
        standard_name="radiation_frequency",
        required=[Product.MWR_L1C],
    ),
    "receiver": Variable(
        long_name="Corresponding microwave receiver for each channel",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "bandwidth": Variable(
        long_name="Bandwidth of microwave channels",
        units="GHz",
        required=[Product.MWR_L1C],
    ),
    "sideband_IF_separation": Variable(
        long_name="Sideband IF separation",
        units="GHz",
        required=[Product.MWR_L1C],
    ),
    "freq_shift": Variable(
        long_name="Frequency shift of the microwave channels",
        units="GHz",
        required=[Product.MWR_L1C],
    ),
    "tb": Variable(
        long_name="Microwave brightness temperature",
        units="K",
        standard_name="brightness_temperature",
        required=[Product.MWR_L1C],
    ),
    "tb_spectrum": Variable(
        long_name="Brightness temperature spectrum",
        units="K",
    ),
    "receiver_nb": Variable(
        long_name="Microwave receiver number",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "met_quality_flag": Variable(
        long_name="Meteorological data quality flag",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "ir_beamwidth": Variable(
        long_name="Beam width of the infrared radiometer",
        units="degree",
    ),
    "n_sidebands": Variable(
        long_name="Number of sidebands",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "pointing_flag": Variable(
        long_name="Pointing flag",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "t_amb": Variable(
        long_name="Ambient target temperature",
        units="K",
        required=[Product.MWR_L1C],
    ),
    "t_rec": Variable(
        long_name="Receiver physical temperature",
        units="K",
        required=[Product.MWR_L1C],
    ),
    "t_sta": Variable(
        long_name="Receiver temperature stability",
        units="K",
        required=[Product.MWR_L1C],
    ),
    "quality_flag_status": Variable(
        long_name="Quality flag status",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "liquid_cloud_flag": Variable(
        long_name="Liquid cloud flag",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "liquid_cloud_flag_status": Variable(
        long_name="Liquid cloud flag status",
        dtype=Dtype.INT,
        required=[Product.MWR_L1C],
    ),
    "ir_wavelength": Variable(
        long_name="Wavelength of infrared channels",
        units="m",
    ),
    "ir_bandwidth": Variable(
        long_name="Bandwidth of infrared channels",
        units="m",
    ),
    "irt": Variable(
        long_name="Infrared brightness temperatures",
        units="K",
    ),
    "ir_azimuth_angle": Variable(
        long_name="Infrared sensor azimuth angle",
        units="degree",
    ),
    "ir_zenith_angle": Variable(
        long_name="Infrared sensor zenith angle",
        units="degree",
    ),
    # ---------------------------
    # Required in MWR SINGLE file
    # ---------------------------
    "iwv_quality_flag": Variable(
        long_name="Quality flag",
        units="1",
        required=[Product.MWR_SINGLE],
        dtype=Dtype.INT,
    ),
    "iwv_quality_flag_status": Variable(
        long_name="Quality flag status",
        units="1",
        required=[Product.MWR_SINGLE],
        dtype=Dtype.INT,
    ),
    "absolute_humidity_quality_flag": Variable(
        long_name="Quality flag",
        units="1",
        required=[Product.MWR_SINGLE],
        dtype=Dtype.INT,
    ),
    "absolute_humidity_quality_flag_status": Variable(
        long_name="Quality flag status",
        units="1",
        required=[Product.MWR_SINGLE],
        dtype=Dtype.INT,
    ),
    "lwp_quality_flag": Variable(
        long_name="Quality flag",
        units="1",
        required=[Product.MWR_SINGLE],
        dtype=Dtype.INT,
    ),
    "lwp_quality_flag_status": Variable(
        long_name="Quality flag status",
        units="1",
        required=[Product.MWR_SINGLE],
        dtype=Dtype.INT,
    ),
    "lwp_offset": Variable(
        long_name="Subtracted offset correction of liquid water path",
        units="kg m-2",
        required=[Product.MWR_SINGLE],
    ),
    "absolute_humidity": Variable(
        long_name="Absolute humidity",
        units="kg m-3",
        required=[Product.MWR_SINGLE],
    ),
    # ----------------------------------------
    # Required in MWR MULTI and MWR MULTI file
    # ----------------------------------------
    "potential_temperature": Variable(
        long_name="Potential temperature",
        units="K",
        required=[
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
        ],
    ),
    "equivalent_potential_temperature": Variable(
        long_name="Equivalent potential temperature",
        units="K",
        required=[
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
        ],
    ),
    "relative_humidity": Variable(
        long_name="Relative humidity",
        standard_name="relative_humidity",
        required=[
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
        ],
    ),
    "temperature_quality_flag": Variable(
        long_name="Quality flag",
        units="1",
        required=[
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
        ],
        dtype=Dtype.INT,
    ),
    "temperature_quality_flag_status": Variable(
        long_name="Quality flag status",
        units="1",
        required=[
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
        ],
        dtype=Dtype.INT,
    ),
    # -------------------------
    # Required in several files
    # -------------------------
    "rainfall_rate": Variable(
        long_name="Rainfall rate",
        units="m s-1",
        standard_name="rainfall_rate",
        required=[
            Product.DISDROMETER,
            Product.WEATHER_STATION,
            Product.RAIN_RADAR,
        ],
    ),
    "range": Variable(
        long_name="Range from instrument",
        units="m",
        required=[
            Product.RADAR,
            Product.LIDAR,
        ],
    ),
    "v": Variable(
        long_name="Doppler velocity",
        units="m s-1",
        required=[
            Product.RADAR,
            Product.CATEGORIZE,
            Product.CATEGORIZE_VOODOO,
            Product.DOPPLER_LIDAR,
            Product.RAIN_RADAR,
        ],
    ),
    "temperature": Variable(
        long_name="Temperature",
        units="K",
        required=[
            Product.MODEL,
            Product.CATEGORIZE,
            Product.CATEGORIZE_VOODOO,
            Product.MWR_MULTI,
        ],
    ),
    "pressure": Variable(
        long_name="Pressure",
        units="Pa",
        required=[
            Product.MODEL,
            Product.CATEGORIZE,
            Product.CATEGORIZE_VOODOO,
        ],
    ),
    "lwp": Variable(
        long_name="Liquid water path",
        units="kg m-2",
        standard_name="atmosphere_cloud_liquid_water_content",
        required=[
            Product.MWR,
            Product.LWC,
            Product.MWR_SINGLE,
        ],
    ),
    "lwp_error": Variable(
        long_name="Error in liquid water path",
        units="kg m-2",
        required=[Product.LWC],
    ),
    "height": Variable(
        long_name="Height above mean sea level",
        units="m",
        standard_name="height_above_mean_sea_level",
        required=Product.all()
        - {
            Product.MWR,
            Product.DISDROMETER,
            Product.WEATHER_STATION,
            Product.MWR_L1C,
            Product.L3_CF,
            Product.L3_IWC,
            Product.L3_LWC,
        },
    ),
    "time": Variable(
        long_name="Time UTC",
        units=time_units,
        standard_name="time",
        required=Product.all(),
    ),
    "altitude": Variable(
        long_name="Altitude of site",
        units="m",
        standard_name="altitude",
        required=Product.all()
        - {Product.MODEL, Product.L3_CF, Product.L3_IWC, Product.L3_LWC},
    ),
    "latitude": Variable(
        long_name="Latitude of site",
        units="degree_north",
        standard_name="latitude",
        required=Product.all(),
    ),
    "longitude": Variable(
        long_name="Longitude of site",
        units="degree_east",
        standard_name="longitude",
        required=Product.all(),
    ),
    # --------------------------------------------
    # Variables included in some of Level 1b files
    # --------------------------------------------
    "calibration_offset": Variable(
        long_name="Radar reflectivity calibration offset",
        units="dBZ",
    ),
    "range_offset": Variable(
        long_name="Radar range offset",
        units="m",
    ),
    "rainfall_rate_1min_total": Variable(
        long_name="Total precipitation rate",
        units="m s-1",
    ),
    "rainfall_rate_1min_solid": Variable(
        long_name="Solid precipitation rate",
        units="m s-1",
    ),
    "snowfall_rate": Variable(
        long_name="Snowfall rate",
        units="m s-1",
    ),
    "velocity": Variable(
        long_name="Center fall velocity of precipitation particles",
        units="m s-1",
    ),
    "velocity_spread": Variable(
        long_name="Width of velocity interval",
        units="m s-1",
    ),
    "velocity_bnds": Variable(
        long_name="Velocity bounds",
        units="m s-1",
    ),
    "diameter": Variable(
        long_name="Center diameter of precipitation particles",
        units="m",
    ),
    "diameter_spread": Variable(
        long_name="Width of diameter interval",
        units="m",
    ),
    "diameter_bnds": Variable(
        long_name="Diameter bounds",
        units="m",
    ),
    "synop_WaWa": Variable(
        long_name="Synop code WaWa",
        dtype=Dtype.INT,
    ),
    "interval": Variable(
        long_name="Length of measurement interval",
        units="s",
        dtype=Dtype.INT,
    ),
    "sig_laser": Variable(
        long_name="Signal amplitude of the laser strip",
        dtype=Dtype.INT,
    ),
    "T_sensor": Variable(
        long_name="Temperature in the sensor housing",
        units="K",
    ),
    "I_heating": Variable(
        long_name="Heating current",
        units="A",
    ),
    "kinetic_energy": Variable(
        long_name="Kinetic energy of the hydrometeors",
        units="J m-2 h-1",
    ),
    "V_power_supply": Variable(
        long_name="Power supply voltage",
        units="V",
    ),
    "V_sensor_supply": Variable(
        long_name="Sensor supply voltage",
        units="V",
    ),
    "state_sensor": Variable(
        long_name="State of the sensor",
        dtype=Dtype.INT,
    ),
    "error_code": Variable(
        long_name="Error code",
        dtype=Dtype.INT,
    ),
    "number_concentration": Variable(
        long_name="Number of particles per diameter class",
        units="m-3 mm-1",
    ),
    "fall_velocity": Variable(
        long_name="Average velocity of each diameter class",
        units="m s-1",
    ),
    "data_raw": Variable(
        long_name="Raw data as a function of particle diameter and velocity",
        units="1",
        dtype=Dtype.SHORT,
    ),
    "phi_cx": Variable(
        long_name="Co-cross-channel differential phase",
        units="rad",
    ),
    "rho_cx": Variable(
        long_name="Co-cross-channel correlation coefficient",
    ),
    "kurtosis": Variable(
        long_name="Kurtosis of spectra",
    ),
    "skewness": Variable(
        long_name="Skewness of spectra",
    ),
    "azimuth_angle": Variable(
        long_name="Azimuth angle",
        units="degree",
        standard_name="sensor_azimuth_angle",
    ),
    "beta_raw": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_cross_raw": Variable(
        long_name="Attenuated backscatter coefficient for the cross-polarised signal",
        units="sr-1 m-1",
    ),
    "beta_cross": Variable(
        long_name="Attenuated backscatter coefficient for the cross-polarised signal",
        units="sr-1 m-1",
    ),
    "iwv": Variable(
        long_name="Integrated water vapour",
        units="kg m-2",
        standard_name="atmosphere_mass_content_of_water_vapor",
    ),
    "ldr": Variable(
        long_name="Linear depolarisation ratio",
        units="dB",
    ),
    "sldr": Variable(
        long_name="Slanted linear depolarisation ratio",
        units="dB",
    ),
    "width": Variable(
        long_name="Spectral width",
        units="m s-1",
        required=[Product.RAIN_RADAR],
    ),
    "calibration_factor": Variable(
        long_name="Attenuated backscatter calibration factor",
    ),
    "beta_smooth": Variable(
        long_name="Attenuated backscatter coefficient", units="sr-1 m-1"
    ),
    "depolarisation": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "depolarisation_raw": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "file_code": Variable(
        long_name="File code",
        dtype=Dtype.INT,
    ),
    "program_number": Variable(
        long_name="Program number",
        dtype=Dtype.INT,
    ),
    "model_number": Variable(
        long_name="Model number",
        dtype=Dtype.INT,
    ),
    "antenna_separation": Variable(
        long_name="Antenna separation",
        units="m",
    ),
    "antenna_diameter": Variable(
        long_name="Antenna diameter",
        units="m",
    ),
    "antenna_gain": Variable(
        long_name="Antenna gain",
        units="dB",
    ),
    "half_power_beam_width": Variable(
        long_name="Half power beam width",
        units="degree",
    ),
    "dual_polarization": Variable(
        long_name="Dual polarisation type",
        dtype=Dtype.INT,
    ),
    "sample_duration": Variable(
        long_name="Sample duration",
        units="s",
    ),
    "calibration_interval": Variable(
        long_name="Calibration interval in samples",
        dtype=Dtype.INT,
    ),
    "number_of_spectral_samples": Variable(
        long_name="Number of spectral samples in each chirp sequence",
        dtype=Dtype.INT,
    ),
    "chirp_start_indices": Variable(
        long_name="Chirp sequences start indices",
        dtype=Dtype.INT,
    ),
    "number_of_averaged_chirps": Variable(
        long_name="Number of averaged chirps in sequence",
        dtype=Dtype.INT,
    ),
    "integration_time": Variable(
        long_name="Integration time",
        units="s",
    ),
    "range_resolution": Variable(
        long_name="Vertical resolution of range",
        units="m",
    ),
    "FFT_window": Variable(
        long_name="FFT window type",
        dtype=Dtype.INT,
    ),
    "input_voltage_range": Variable(
        long_name="ADC input voltage range (+/-)",
        units="mV",
        dtype=Dtype.INT,
    ),
    "noise_threshold": Variable(
        long_name="Noise filter threshold factor",
    ),
    "time_ms": Variable(
        long_name="Time ms",
        units="ms",
        dtype=Dtype.INT,
    ),
    "quality_flag": Variable(
        long_name="Quality flag",
        dtype=Dtype.INT,
    ),
    "pc_temperature": Variable(
        long_name="PC temperature",
        units="K",
    ),
    "receiver_temperature": Variable(
        long_name="Receiver temperature",
        units="K",
    ),
    "transmitter_temperature": Variable(
        long_name="Transmitter temperature",
        units="K",
    ),
    "transmitted_power": Variable(
        long_name="Transmitted power",
        units="W",
    ),
    "status_flag": Variable(
        long_name="Status flag for heater and blower",
    ),
    "wind_speed": Variable(
        long_name="Wind speed",
        standard_name="wind_speed",
        units="m s-1",
    ),
    "wind_direction": Variable(
        long_name="Wind direction",
        units="degree",
        standard_name="wind_from_direction",
    ),
    "voltage": Variable(
        long_name="Voltage",
        units="V",
    ),
    "brightness_temperature": Variable(
        long_name="Brightness temperature",
        units="K",
    ),
    "if_power": Variable(
        long_name="IF power at ACD",
        units="uW",
    ),
    "level": Variable(
        long_name="Model level",
        units=None,
        dtype=Dtype.SHORT,
    ),
    "flux_level": Variable(
        long_name="Model flux level",
        units=None,
        dtype=Dtype.SHORT,
    ),
    "sfc_categorical_snow": Variable(
        long_name="",
        dtype=Dtype.SHORT,
    ),
    "sfc_categorical_ice": Variable(
        long_name="",
        dtype=Dtype.SHORT,
    ),
    "sfc_categorical_freezing_rain": Variable(
        long_name="",
        dtype=Dtype.SHORT,
    ),
    "sfc_categorical_rain": Variable(
        long_name="",
        dtype=Dtype.SHORT,
    ),
    "sfc_albedo_sw_direct": Variable(
        long_name="Surface albedo (shortwave direct)",
        # standard_name="surface_albedo_shortwave_direct",
    ),
    "sfc_albedo_sw_diffuse": Variable(
        long_name="Surface albedo (shortwave diffuse)",
        # standard_name="surface_albedo_shortwave_diffuse",
    ),
    "sfc_albedo_lw_direct": Variable(
        long_name="Surface albedo (longwave direct)",
        # standard_name="surface_albedo_longwave_direct",
    ),
    "sfc_albedo_lw_diffuse": Variable(
        long_name="Surface albedo (longwave diffuse)",
        # standard_name="surface_albedo_longwave_diffuse",
    ),
    "nfft": Variable(
        long_name="Number of FFT points",
        dtype=Dtype.INT,
    ),
    "nave": Variable(
        long_name="Number of spectral averages (not accounting for overlapping FFTs)",
        dtype=Dtype.INT,
    ),
    "prf": Variable(
        long_name="Pulse Repetition Frequency",
        units="Hz",
        dtype=Dtype.INT,
    ),
    "rg0": Variable(
        long_name="Number of lowest range gates",
        dtype=Dtype.INT,
    ),
    "SNR": Variable(
        long_name="Signal-to-noise ratio",
        units="dB",
    ),
    "radar_pitch": Variable(
        long_name="Radar pitch angle",
        units="degree",
        standard_name="platform_roll",
    ),
    "radar_yaw": Variable(
        long_name="Radar yaw angle",
        units="degree",
        standard_name="platform_yaw",
    ),
    "radar_roll": Variable(
        long_name="Radar roll angle",
        units="degree",
        standard_name="platform_roll",
    ),
    "zdr": Variable(
        long_name="Differential reflectivity",
        units="dB",
    ),
    "rho_hv": Variable(
        long_name="Correlation coefficient",
    ),
    "phi_dp": Variable(
        long_name="Differential phase",
        units="rad",
    ),
    "srho_hv": Variable(
        long_name="Slanted correlation coefficient",
    ),
    "kdp": Variable(
        long_name="Specific differential phase shift",
        units="rad km-1",
    ),
    "differential_attenuation": Variable(
        long_name="Differential attenuation",
        units="dB km-1",
    ),
    "synop_WW": Variable(
        long_name="Synop code WW",
        dtype=Dtype.INT,
    ),
    "measurement_quality": Variable(
        long_name="Measurement quality",
        units="%",
        dtype=Dtype.INT,
    ),
    "status_laser": Variable(
        long_name="Status of laser",
        dtype=Dtype.INT,
    ),
    "static_signal": Variable(
        long_name="Static signal",
        dtype=Dtype.INT,
    ),
    "status_T_laser_analogue": Variable(
        long_name="Status of laser temperature (analogue)",
        dtype=Dtype.INT,
    ),
    "status_T_laser_digital": Variable(
        long_name="Status of laser temperature (digital)",
        dtype=Dtype.INT,
    ),
    "status_I_laser_analogue": Variable(
        long_name="Status of laser current (analogue)",
        dtype=Dtype.INT,
    ),
    "status_I_laser_digital": Variable(
        long_name="Status of laser current (digital)",
        dtype=Dtype.INT,
    ),
    "status_sensor_supply": Variable(
        long_name="Status of sensor supply",
        dtype=Dtype.INT,
    ),
    "status_laser_heating": Variable(
        long_name="Status of laser heating",
        dtype=Dtype.INT,
    ),
    "status_receiver_heating": Variable(
        long_name="Status of receiver heating",
        dtype=Dtype.INT,
    ),
    "status_temperature_sensor": Variable(
        long_name="Status of temperature sensor",
        dtype=Dtype.INT,
    ),
    "status_heating_supply": Variable(
        long_name="Status of heating supply",
        dtype=Dtype.INT,
    ),
    "status_heating_housing": Variable(
        long_name="Status of heating housing",
        dtype=Dtype.INT,
    ),
    "status_heating_heads": Variable(
        long_name="Status of heating heads",
        dtype=Dtype.INT,
    ),
    "status_heating_carriers": Variable(
        long_name="Status of heating carriers",
        dtype=Dtype.INT,
    ),
    "status_laser_power": Variable(
        long_name="Status of laser power",
        dtype=Dtype.INT,
    ),
    "T_interior": Variable(
        long_name="Interior temperature",
        units="K",
    ),
    "T_laser_driver": Variable(
        long_name="Temperature of laser driver",
        units="K",
    ),
    "T_ambient": Variable(
        long_name="Ambient temperature",
        units="K",
    ),
    "I_mean_laser": Variable(
        long_name="Mean value of laser current",
        units="mA",
    ),
    "V_control": Variable(
        long_name="Control voltage",
        units="mV",
        dtype=Dtype.INT,
    ),
    "V_optical_output": Variable(
        long_name="Voltage of optical control output",
        units="mV",
        dtype=Dtype.INT,
    ),
    "I_heating_laser_head": Variable(
        long_name="Laser head heating current",
        units="mA",
        dtype=Dtype.INT,
    ),
    "I_heating_receiver_head": Variable(
        long_name="Receiver head heating current",
        units="mA",
        dtype=Dtype.INT,
    ),
    "maximum_hail_diameter": Variable(
        long_name="Maximum hail diameter",
        units="mm",
    ),
    # rv-polarstern specific
    "beta_1064_raw": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_1064": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_532_raw": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_532": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_355_raw": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_355": Variable(
        long_name="Attenuated backscatter coefficient",
        units="sr-1 m-1",
    ),
    "beta_355_nr": Variable(
        long_name="Attenuated backscatter coefficient (near range)",
        units="sr-1 m-1",
    ),
    "beta_355_nr_raw": Variable(
        long_name="Attenuated backscatter coefficient (near range)",
        units="sr-1 m-1",
    ),
    "beta_532_nr": Variable(
        long_name="Attenuated backscatter coefficient (near range)",
        units="sr-1 m-1",
    ),
    "beta_532_nr_raw": Variable(
        long_name="Attenuated backscatter coefficient (near range)",
        units="sr-1 m-1",
    ),
    "calibration_factor_1064": Variable(
        long_name="Attenuated backscatter at 1064 calibration factor",
    ),
    "calibration_factor_532": Variable(
        long_name="Attenuated backscatter at 532 calibration factor",
    ),
    "calibration_factor_355": Variable(
        long_name="Attenuated backscatter at 355 calibration factor",
    ),
    "calibration_factor_355_nr": Variable(
        long_name="Attenuated backscatter at 355 near range calibration factor",
    ),
    "calibration_factor_532_nr": Variable(
        long_name="Attenuated backscatter at 532 near range calibration factor",
    ),
    "depolarisation_355_raw": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "depolarisation_355": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "depolarisation_532_raw": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "depolarisation_532": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "depolarisation_1064": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "depolarisation_1064_raw": Variable(
        long_name="Lidar volume linear depolarisation ratio",
    ),
    "snr_1064": Variable(
        long_name="Signal-to-Noise Ratio (1064 nm)",
    ),
    "snr_532": Variable(
        long_name="Signal-to-Noise Ratio (532 nm)",
    ),
    "snr_355": Variable(
        long_name="Signal-to-Noise Ratio (355 nm)",
    ),
    "snr_532_nr": Variable(
        long_name="Signal-to-Noise Ratio (532 nm near range)",
    ),
    "snr_355_nr": Variable(
        long_name="Signal-to-Noise Ratio (355 nm near range)",
    ),
    "lidar_depolarisation": Variable(
        long_name="Calibrated volume depolarisation (532 nm)"  # in rv-polarstern categorize files
    ),
    "beamwidthV": Variable(
        long_name="Vertical angular beamwidth",
        units="degree",
    ),
    "beamwidthH": Variable(
        long_name="Horizontal angular beamwidth",
        units="degree",
    ),
    "rain_detected": Variable(long_name="Rain detected", dtype=Dtype.INT),
    "air_pressure_quality_flag": Variable(
        long_name="Air pressure quality flag",
        dtype=Dtype.BYTE,
    ),
    "air_temperature_quality_flag": Variable(
        long_name="Air temperature quality flag",
        dtype=Dtype.BYTE,
    ),
    "relative_humidity_quality_flag": Variable(
        long_name="Relative humidity quality flag",
        dtype=Dtype.BYTE,
    ),
    "wind_speed_quality_flag": Variable(
        long_name="Wind speed quality flag",
        dtype=Dtype.BYTE,
    ),
    "wind_direction_quality_flag": Variable(
        long_name="Wind direction quality flag",
        dtype=Dtype.BYTE,
    ),
    "rainfall_rate_quality_flag": Variable(
        long_name="Rainfall rate quality flag",
        dtype=Dtype.BYTE,
    ),
    "dew_point_temperature": Variable(
        long_name="Dew point temperature",
        dtype=Dtype.FLOAT,
        units="K",
    ),
    "dew_point_temperature_quality_flag": Variable(
        long_name="Dew point temperature quality flag",
        dtype=Dtype.BYTE,
    ),
    "wind_speed_gust": Variable(
        long_name="Wind speed gust",
        dtype=Dtype.FLOAT,
        units="m s-1",
    ),
    "wind_speed_gust_quality_flag": Variable(
        long_name="Wind speed gust quality flag",
        dtype=Dtype.BYTE,
    ),
    "visibility_quality_flag": Variable(
        long_name="Visibility quality flag",
        dtype=Dtype.BYTE,
    ),
    "elevation_angle": Variable(
        long_name="Sensor elevation angle",
        units="degree",
    ),
    "ir_elevation_angle": Variable(
        long_name="Infrared sensor elevation angle",
        units="degree",
    ),
    "liquid_prob": Variable(
        long_name="Liquid probability",
        required=[Product.CATEGORIZE_VOODOO],
    ),
    "polariser_bleed_through": Variable(
        long_name="Polariser bleed-through",
    ),
}



import pytest
from numpy.testing import assert_allclose

from cloudnetpy_qc.utils import calc_pressure


@pytest.mark.parametrize(
    "altitude,pressure",
    [
        (0, 101325),
        (1036.3, 89479),
        (2011.7, 79380),
        (3048.0, 69682),
        (4023.4, 61453),
        (5059.7, 53590),
    ],
)
def test_calc_pressure(altitude, pressure):
    assert_allclose(calc_pressure(altitude), pressure, rtol=1e-5)



"""Helper functions."""

import configparser
import re
from functools import lru_cache

import numpy as np
import requests

PID_FORMAT = r"https?://hdl\.handle\.net/(.+)"


def read_config(filename: str) -> configparser.ConfigParser:
    conf = configparser.ConfigParser()
    conf.optionxform = str  # type: ignore
    conf.read(filename)
    return conf


def format_msg(msg_in: str | list) -> str:
    msg = msg_in[0] if isinstance(msg_in, list) else msg_in
    if not msg.endswith("."):
        msg += "."
    x = re.search("^\\(.+\\):", msg)
    if x:
        msg = msg[x.end() :]
    msg = re.sub(" +", " ", msg)
    msg = msg.strip()
    msg = msg[0].capitalize() + msg[1:]
    return msg


def format_list(values: list[str], conj: str) -> str:
    if len(values) == 0:
        return ""
    if len(values) == 1:
        return values[0]
    return ", ".join(values[:-1]) + f" {conj} " + values[-1]


def create_expected_received_msg(
    expected: str | list[str],
    received: str,
    variable: str | None = None,
) -> str:
    if isinstance(expected, str):
        expected = [expected]
    expected = format_list([f"'{var}'" for var in expected], "or")
    msg = f"Expected {expected} but received '{received}'"
    if variable is not None:
        return f"{msg} with variable '{variable}'"
    return msg


def create_out_of_bounds_msg(
    variable: str,
    lower_limit: str | int | float,
    upper_limit: str | int | float,
    value: str | int | float,
) -> str:
    return (
        f"Value {format_value(value)} exceeds expected limits {format_value(lower_limit)} ... "
        f"{format_value(upper_limit)} with variable '{variable}'"
    )


def format_value(value: str | int | float) -> str:
    return "{:,g}".format(float(value)).replace(",", "\u202f")


@lru_cache
def fetch_pid(pid: str) -> dict:
    match = re.fullmatch(PID_FORMAT, pid)
    if match is None:
        raise ValueError("Invalid PID format")
    url = "https://hdl.handle.net/api/handles/" + match[1]
    res = requests.get(url, timeout=30)
    res.raise_for_status()
    return res.json()


def integer_ranges(ints: list[int]) -> list[str]:
    """
    Convert given integers to list of ranges.
    >>> integer_ranges([1,2,3,5,7,8,9])
    ['1–3', '5', '7–9']
    """
    if len(ints) == 0:
        return []
    ints = sorted(ints)
    output = [[ints[0], ints[0]]]
    for x in ints[1:]:
        if x == output[-1][1] + 1:
            output[-1][1] = x
        else:
            output.append([x, x])
    return [str(a) if a == b else f"{a}–{b}" for a, b in output]


def calc_pressure(altitude: float) -> float:
    """
    Calculate atmospheric pressure in International Standard Atmosphere.

    Args:
        altitude: Geopotential altitude above mean sea level (m)

    Returns:
        Atmospheric pressure (Pa)
    """
    if np.any(altitude >= 11_000):
        raise NotImplementedError("Implemented only up to 11 km")
    P0 = 101_325  # Standard atmospheric pressure at sea level (Pa)
    T0 = 288.15  # Standard temperature at sea level (K)
    L = 0.0065  # Temperature lapse rate (K m-1)
    g = 9.80665  # Gravitational acceleration (m s-2)
    R = 287.0528  # Specific gas constant (J kg-1 K-1)
    return P0 * (1 - L * altitude / T0) ** (g / (R * L))



"""cloudnetpy-qc tests."""

from os import path
from pathlib import Path

from cloudnetpy_qc import quality

SCRIPT_PATH = path.dirname(path.realpath(__file__))


def test_valid_file():
    # TODO: update this file
    filename = f"{SCRIPT_PATH}/data/20211129_juelich_hatpro.nc"
    check = Check(filename)
    check.errors()
    for test in check.tests:
        excs = test.exceptions
        assert (
            excs
            if test.test_id
            in (
                "TestGlobalAttributes",
                "TestInstrumentPid",
                "TestUnits",
                "FindVariableOutliers",
            )
            else not excs
        ), f"{test}, {excs}"


def test_legacy_file():
    filename = f"{SCRIPT_PATH}/data/20120203_arm-maldives_classification.nc"
    check = Check(filename, file_type="classification")
    keys = [
        "TestLongNames",
        "TestStandardNames",
        "TestDataTypes",
        "TestGlobalAttributes",
        "TestVariableNames",
        "TestCFConvention",
    ]
    check.verify_exceptions(keys)


def test_missing_data():
    filename = f"{SCRIPT_PATH}/data/20220729_norunda_cl51.nc"
    check = Check(filename)
    check.verify_exceptions(["TestDataCoverage"])


def test_invalid_lwp():
    filename = f"{SCRIPT_PATH}/data/20220214_schneefernerhaus_hatpro.nc"
    check = Check(filename)
    check.verify_exceptions(["FindVariableOutliers"])


def test_file_without_time_array():
    filename = f"{SCRIPT_PATH}/data/20200505_chilbolton_mira.nc"
    check = Check(filename)
    check.verify_exceptions(["TestTimeVector"])


def test_bad_mwr_from_delft():
    filename = f"{SCRIPT_PATH}/data/20210421_delft_hatpro.nc"
    check = Check(filename)
    check.verify_exceptions(["TestTimeVector"])


def test_bad_mwr_from_granada():
    filename = f"{SCRIPT_PATH}/data/20160610_granada_hatpro.nc"
    check = Check(filename)
    check.verify_exceptions(["TestTimeVector"])


def test_bad_categorize_from_chilbolton():
    filename = f"{SCRIPT_PATH}/data/20001017_chilbolton_categorize.nc"
    check = Check(filename, file_type="categorize")
    check.verify_exceptions(["FindFolding"])


def test_empty_instrument_pid():
    filename = f"{SCRIPT_PATH}/data/20220326_schneefernerhaus_mira.nc"
    check = Check(filename)
    check.verify_exceptions(["TestInstrumentPid"])


class Check:
    """Check class."""

    def __init__(self, filename: str, file_type: str | None = None):
        self.report = quality.run_tests(Path(filename), product=file_type)
        self.tests = self.report.tests

    def verify_exceptions(self, keys: list):
        n = 0
        for test in self.tests:
            if test.test_id in keys:
                assert test.exceptions
                n += 1
        assert n == len(keys)

    def errors(self, n_expected: int = 0):
        assert self._count("error") == n_expected

    def warnings(self, n_expected: int = 0):
        assert self._count("warning") == n_expected

    def _count(self, level: str):
        n = 0
        for test in self.report.tests:
            for exc in test.exceptions:
                if exc.result == level:
                    n += 1
        return n



"""Cloudnet product quality checks."""

import datetime
import json
import logging
import os
import re
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Iterable, NamedTuple

import netCDF4
import numpy as np
import scipy.stats
from numpy import ma
from requests import RequestException

from . import utils
from .variables import LEVELS, VARIABLES, Product
from .version import __version__

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")

METADATA_CONFIG = utils.read_config(os.path.join(DATA_PATH, "metadata_config.ini"))
DATA_CONFIG = utils.read_config(os.path.join(DATA_PATH, "data_quality_config.ini"))
CF_AREA_TYPES_XML = os.path.join(DATA_PATH, "area-type-table.xml")
CF_STANDARD_NAMES_XML = os.path.join(DATA_PATH, "cf-standard-name-table.xml")
CF_REGION_NAMES_XML = os.path.join(DATA_PATH, "standardized-region-list.xml")


class ErrorLevel(Enum):
    WARNING = "warning"
    ERROR = "error"
    INFO = "info"


class TestException(NamedTuple):
    result: ErrorLevel
    message: str


class TestReport(NamedTuple):
    test_id: str
    exceptions: list[TestException]


class FileReport(NamedTuple):
    timestamp: datetime.datetime
    qc_version: str
    tests: list[TestReport]

    def to_dict(self) -> dict:
        return {
            "timestamp": self.timestamp.isoformat(),
            "qcVersion": self.qc_version,
            "tests": [
                {
                    "testId": test.test_id,
                    "exceptions": [
                        {"result": exception.result.value, "message": exception.message}
                        for exception in test.exceptions
                    ],
                }
                for test in self.tests
            ],
        }


def run_tests(
    filename: str | PathLike,
    product: Product | str | None = None,
    ignore_tests: list[str] | None = None,
) -> FileReport:
    filename = Path(filename)
    if isinstance(product, str):
        product = Product(product)
    with netCDF4.Dataset(filename) as nc:
        if product is None:
            try:
                product = Product(nc.cloudnet_file_type)
            except AttributeError as exc:
                raise ValueError(
                    "No 'cloudnet_file_type' global attribute found, can not run tests. "
                    "Is this a legacy file?"
                ) from exc
        logging.debug(f"Filename: {filename.stem}")
        logging.debug(f"File type: {product}")
        test_reports: list[TestReport] = []
        for cls in Test.__subclasses__():
            if ignore_tests and cls.__name__ in ignore_tests:
                continue
            test_instance = cls(nc, filename, product)
            if product not in test_instance.products:
                continue
            try:
                test_instance.run()
            except Exception as err:
                test_instance._add_error(
                    f"Failed to run test: {err} ({type(err).__name__})"
                )
            test_reports.append(test_instance.report)
    return FileReport(
        timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
        qc_version=__version__,
        tests=test_reports,
    )


class Test:
    """Test base class."""

    name: str
    description: str
    products: Iterable[Product] = Product.all()

    def __init__(self, nc: netCDF4.Dataset, filename: Path, product: Product):
        self.filename = filename
        self.nc = nc
        self.product = product
        self.report = TestReport(
            test_id=self.__class__.__name__,
            exceptions=[],
        )

    def run(self):
        raise NotImplementedError

    def _add_message(self, message: str | list, severity: ErrorLevel):
        self.report.exceptions.append(
            TestException(result=severity, message=utils.format_msg(message))
        )

    def _add_info(self, message: str | list):
        self._add_message(message, ErrorLevel.INFO)

    def _add_warning(self, message: str | list):
        self._add_message(message, ErrorLevel.WARNING)

    def _add_error(self, message: str | list):
        self._add_message(message, ErrorLevel.ERROR)

    def _read_config_keys(self, config_section: str) -> np.ndarray:
        field = "all" if "attr" in config_section else self.product.value
        keys = METADATA_CONFIG[config_section][field].split(",")
        return np.char.strip(keys)

    def _get_required_variables(self) -> dict:
        return {
            name: var
            for name, var in VARIABLES.items()
            if var.required is not None and self.product in var.required
        }

    def _get_required_variable_names(self) -> set:
        required_variables = self._get_required_variables()
        return set(required_variables.keys())

    def _test_variable_attribute(self, attribute: str):
        for key in self.nc.variables.keys():
            if key not in VARIABLES:
                continue
            expected = getattr(VARIABLES[key], attribute)
            if callable(expected):
                expected = expected(self.nc)
            if expected is not None:
                value = getattr(self.nc.variables[key], attribute, "")
                if value != expected:
                    msg = utils.create_expected_received_msg(
                        expected, value, variable=key
                    )
                    self._add_warning(msg)

    def _get_date(self):
        date_in_file = [int(getattr(self.nc, x)) for x in ("year", "month", "day")]
        return datetime.date(*date_in_file)

    def _get_duration(self) -> datetime.timedelta:
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        if now.date() == self._get_date():
            midnight = now.replace(hour=0, minute=0, second=0, microsecond=0)
            duration = now - midnight
        else:
            duration = datetime.timedelta(days=1)
        return duration


# --------------------#
# ------ Infos ------ #
# --------------------#


class FindVariableOutliers(Test):
    name = "Variable outliers"
    description = "Find suspicious data values."

    def run(self):
        for key in self.nc.variables:
            limits = self._get_limits(key)
            if limits is None:
                continue
            data = self.nc.variables[key][:]
            if data.ndim > 0 and len(data) == 0:
                break
            max_value = np.max(data)
            min_value = np.min(data)
            if min_value < limits[0]:
                msg = utils.create_out_of_bounds_msg(key, *limits, min_value)
                self._add_info(msg)
            if max_value > limits[1]:
                msg = utils.create_out_of_bounds_msg(key, *limits, max_value)
                self._add_info(msg)

    def _get_limits(self, key: str) -> tuple[float, float] | None:
        if key == "zenith_angle" and self.product in (
            Product.MWR_L1C,
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
        ):
            return None
        if key == "air_pressure":
            pressure = utils.calc_pressure(self.nc["altitude"][:])
            max_diff = pressure * 0.05
            return (pressure - max_diff, pressure + max_diff)
        if not DATA_CONFIG.has_option("limits", key):
            return None
        limit_min, limit_max = DATA_CONFIG.get("limits", key).split(",", maxsplit=1)
        return (float(limit_min), float(limit_max))


class FindFolding(Test):
    name = "Radar folding"
    description = "Test for radar folding."
    products = [Product.RADAR, Product.CATEGORIZE]

    def run(self):
        key = "v"
        v_threshold = 8
        try:
            data = self.nc[key][:]
        except IndexError:
            self._add_error(f"Doppler velocity, '{key}', is missing.")
            return
        difference = np.abs(np.diff(data, axis=1))
        n_suspicious = ma.sum(difference > v_threshold)
        if n_suspicious > 20:
            self._add_info(
                f"{n_suspicious} suspicious pixels. Folding might be present."
            )


class TestDataCoverage(Test):
    name = "Data coverage"
    description = "Test that file contains enough data."

    RESOLUTIONS = {
        Product.DISDROMETER: datetime.timedelta(minutes=1),
        Product.L3_CF: datetime.timedelta(hours=1),
        Product.L3_IWC: datetime.timedelta(hours=1),
        Product.L3_LWC: datetime.timedelta(hours=1),
        Product.MWR: datetime.timedelta(minutes=5),
        Product.MWR_MULTI: datetime.timedelta(minutes=30),
        Product.MWR_SINGLE: datetime.timedelta(minutes=5),
        Product.WEATHER_STATION: datetime.timedelta(minutes=10),
        Product.DOPPLER_LIDAR_WIND: datetime.timedelta(hours=1.5),
    }
    DEFAULT_RESOLUTION = datetime.timedelta(seconds=30)

    def _model_resolution(self):
        source = self.nc.source.lower()
        if "gdas" in source or "ecmwf open" in source:
            return datetime.timedelta(hours=3)
        return datetime.timedelta(hours=1)

    def run(self):
        time = np.array(self.nc["time"][:])
        time_unit = datetime.timedelta(hours=1)
        try:
            n_time = len(time)
        except (TypeError, ValueError):
            return
        if n_time < 2:
            return
        if self.nc.cloudnet_file_type == "model":
            expected_res = self._model_resolution()
        else:
            expected_res = self.RESOLUTIONS.get(self.product, self.DEFAULT_RESOLUTION)
        duration = self._get_duration()
        bins = max(1, duration // expected_res)
        hist, _bin_edges = np.histogram(
            time, bins=bins, range=(0, duration / time_unit)
        )
        missing = np.count_nonzero(hist == 0) / len(hist) * 100
        if missing > 20:
            message = f"{round(missing)}% of day's data is missing."
            if missing > 60:
                self._add_warning(message)
            else:
                self._add_info(message)

        actual_res = np.median(np.diff(time)) * time_unit
        if actual_res > expected_res * 1.05:
            self._add_warning(
                f"Expected a measurement with interval at least {expected_res},"
                f" got {actual_res} instead"
            )


class TestVariableNamesDefined(Test):
    name = "Variable names"
    description = "Check that variables have expected names."
    products = Product.all() - {
        Product.MODEL,
        Product.L3_CF,
        Product.L3_IWC,
        Product.L3_LWC,
    }

    def run(self):
        for key in self.nc.variables.keys():
            if key not in VARIABLES:
                self._add_info(f"'{key}' is not defined in cloudnetpy-qc.")


# ---------------------- #
# ------ Warnings ------ #
# ---------------------- #


class TestUnits(Test):
    name = "Units"
    description = "Check that variables have expected units."

    def run(self):
        self._test_variable_attribute("units")


class TestLongNames(Test):
    name = "Long names"
    description = "Check that variables have expected long names."
    products = Product.all() - {
        Product.MODEL,
        Product.L3_CF,
        Product.L3_IWC,
        Product.L3_LWC,
    }

    def run(self):
        self._test_variable_attribute("long_name")


class TestStandardNames(Test):
    name = "Standard names"
    description = "Check that variable have expected standard names."
    products = Product.all() - {
        Product.MODEL,
        Product.L3_CF,
        Product.L3_IWC,
        Product.L3_LWC,
    }

    def run(self):
        self._test_variable_attribute("standard_name")


class TestDataTypes(Test):
    name = "Data types"
    description = "Check that variables have expected data types."

    def run(self):
        for key in self.nc.variables:
            if key not in VARIABLES:
                continue
            expected = VARIABLES[key].dtype.value
            received = self.nc.variables[key].dtype.name
            if received != expected:
                if key == "time" and received in ("float32", "float64"):
                    continue
                msg = utils.create_expected_received_msg(
                    expected, received, variable=key
                )
                self._add_warning(msg)


# class TestTimeVectorDataType(Test):
#     name = "Time data type"
#     description = "Check that time vector is in double precision."
#     products=[
#         Product.RADAR,
#         Product.LIDAR,
#         Product.MWR,
#         Product.DISDROMETER,
#         Product.WEATHER_STATION,
#     ]
#     def run(self):
#         key = "time"
#         received = self.nc.variables[key].dtype.name
#         expected = "float64"
#         if received != expected:
#             msg = utils.create_expected_received_msg(key, expected, received)
#             self._add_warning(msg)


class TestGlobalAttributes(Test):
    name = "Global attributes"
    description = "Check that file contains required global attributes."

    REQUIRED_ATTRS = {
        "year",
        "month",
        "day",
        "file_uuid",
        "Conventions",
        "location",
        "history",
        "title",
        "cloudnet_file_type",
        "source",
    }

    def _instrument_product(self, product: Product):
        return (LEVELS[product] == "1b" and product != Product.MODEL) or product in (
            Product.MWR_L1C,
            Product.MWR_SINGLE,
            Product.MWR_MULTI,
            Product.DOPPLER_LIDAR_WIND,
        )

    def _required_attrs(self, product: Product):
        attrs = set(self.REQUIRED_ATTRS)
        if product == Product.MWR_L1C:
            attrs.add("mwrpy_coefficients")
        if product in (Product.MWR_SINGLE, Product.MWR_MULTI):
            attrs.add("source_file_uuids")
        if product != Product.MODEL:
            attrs.add(
                "instrument_pid"
                if self._instrument_product(product)
                else "source_file_uuids"
            )
        return attrs

    def _optional_attr(self, name: str, product: Product) -> bool:
        return (
            name in ("references", "pid")
            or name.endswith("_version")
            or (
                product == Product.MODEL
                and name in ("initialization_time", "institution")
            )
            or (self._instrument_product(product) and name == "serial_number")
        )

    def run(self):
        nc_keys = set(self.nc.ncattrs())
        required_attrs = self._required_attrs(self.product)
        missing_keys = required_attrs - nc_keys
        for key in missing_keys:
            self._add_warning(f"Attribute '{key}' is missing.")
        extra_keys = nc_keys - required_attrs
        for key in extra_keys:
            if not self._optional_attr(key, self.product):
                self._add_warning(f"Unknown attribute '{key}' found.")


class TestMedianLwp(Test):
    name = "Median LWP"
    description = "Test that LWP data are valid."
    products = [Product.MWR, Product.CATEGORIZE]

    def run(self):
        key = "lwp"
        if key not in self.nc.variables:
            self._add_warning(f"'{key}' is missing.")
            return
        data = self.nc.variables[key][:]
        mask_percentage = ma.count_masked(data) / data.size * 100
        if mask_percentage > 20:
            msg = (
                f"{round(mask_percentage,1)} % of '{key}' data points are masked "
                f"due to low quality data."
            )
            if mask_percentage > 60:
                self._add_warning(msg)
            else:
                self._add_info(msg)
        limits = [-0.5, 10]
        median_lwp = ma.median(data) / 1000  # g -> kg
        if median_lwp < limits[0] or median_lwp > limits[1]:
            msg = utils.create_out_of_bounds_msg(key, *limits, median_lwp)
            self._add_warning(msg)


class FindAttributeOutliers(Test):
    name = "Attribute outliers"
    description = "Find suspicious values in global attributes."

    def run(self):
        try:
            year = int(self.nc.year)
            month = int(self.nc.month)
            day = int(self.nc.day)
            datetime.date(year, month, day)
        except AttributeError:
            self._add_warning("Missing some date attributes.")
        except ValueError:
            self._add_warning("Invalid date attributes.")


class TestLDR(Test):
    name = "LDR values"
    description = "Test that LDR values are proper."
    products = [Product.RADAR, Product.CATEGORIZE]

    def run(self):
        has_ldr = "ldr" in self.nc.variables or "sldr" in self.nc.variables
        has_v = "v" in self.nc.variables
        if has_v and has_ldr:
            v = self.nc["v"][:]
            ldr = (
                self.nc["ldr"][:] if "ldr" in self.nc.variables else self.nc["sldr"][:]
            )
            v_count = ma.count(v)
            ldr_count = ma.count(ldr)
            if v_count > 0 and (ldr_count / v_count * 100) < 0.1:
                self._add_warning("LDR exists in less than 0.1 % of pixels.")


class TestMask(Test):
    name = "Data mask"
    description = "Test that data are not completely masked."
    products = [Product.RADAR]

    def run(self):
        if not np.any(~self.nc["v"][:].mask):
            self._add_error("All data are masked.")


class TestIfRangeCorrected(Test):
    name = "Range correction"
    description = "Test that beta is range corrected."
    products = [Product.LIDAR]

    def run(self):
        try:
            range_var = self.nc["range"]
            beta_raw = self.nc["beta_raw"]
        except IndexError:
            return

        n_top_ranges = len(range_var) // 2
        x = range_var[-n_top_ranges:] ** 2
        y = np.std(beta_raw[:, -n_top_ranges:], axis=0)
        sgl_res = scipy.stats.siegelslopes(y, x)
        residuals = np.abs(y - (sgl_res.intercept + sgl_res.slope * x))
        outliers = residuals > 20 * np.percentile(
            residuals, 25
        )  # Ad hoc outlier detection
        res = scipy.stats.pearsonr(x[~outliers], y[~outliers])
        if res.statistic < 0.75:
            self._add_warning("Data might not be range corrected.")


class TestFloatingPointValues(Test):
    name = "Floating-point values"
    description = (
        "Test for special floating-point values "
        "which may indicate problems with the processing."
    )

    def run(self):
        for name, variable in self.nc.variables.items():
            if variable.dtype.kind != "f":
                continue
            if np.isnan(variable[:]).any():
                self._add_warning(f"Variable '{name}' contains NaN value(s).")
            if np.isinf(variable[:]).any():
                self._add_warning(f"Variable '{name}' contains infinite value(s).")


class TestFillValue(Test):
    name = "Fill value"
    description = (
        "Test that fill value is explicitly set for variables with missing data."
    )

    def run(self):
        for name, variable in self.nc.variables.items():
            if variable[:].mask.any() and not hasattr(variable, "_FillValue"):
                self._add_warning(
                    f"Attribute '_FillValue' is missing from variable '{name}'."
                )


# ---------------------#
# ------ Errors ------ #
# -------------------- #


class TestBrightnessTemperature(Test):
    name = "Brightness temperature"
    description = "Test that brightness temperature data are valid."
    products = [Product.MWR_L1C]

    def run(self):
        flags = self.nc["quality_flag"][:]
        bad_percentage = ma.sum(flags != 0) / flags.size * 100
        if bad_percentage > 90:
            self._add_error("More than 90% of the data are flagged.")
        elif bad_percentage > 50:
            self._add_warning("More than 50% of the data are flagged.")


class TestMWRSingleLWP(Test):
    name = "MWR single pointing LWP"
    description = "Test that LWP data are valid."
    products = [Product.MWR_SINGLE]

    def run(self):
        flags = self.nc["lwp_quality_flag"][:]
        bad_percentage = ma.sum(flags != 0) / flags.size * 100
        if bad_percentage > 90:
            self._add_error("More than 90% of the data are flagged.")
        elif bad_percentage > 50:
            self._add_warning("More than 50% of the data are flagged.")


class TestMWRMultiTemperature(Test):
    name = "MWR multiple pointing temperature"
    description = "Test that temperature data are valid."
    products = [Product.MWR_MULTI]

    def run(self):
        flags = self.nc["temperature_quality_flag"][:]
        if not np.any(flags == 0):
            self._add_error("No valid temperature data found.")


class TestLidarBeta(Test):
    name = "Beta presence"
    description = "Test that one beta variable exists."
    products = [Product.LIDAR]

    def run(self):
        valid_keys = {"beta", "beta_1064", "beta_532", "beta_355"}
        for key in valid_keys:
            if key in self.nc.variables:
                return
        self._add_error("No valid beta variable found.")


class TestTimeVector(Test):
    name = "Time vector"
    description = "Test that time vector is continuous."

    def run(self):
        time = self.nc["time"][:]
        try:
            n_time = len(time)
        except (TypeError, ValueError):
            self._add_error("Time vector is empty.")
            return
        if n_time == 0:
            self._add_error("Time vector is empty.")
            return
        if n_time == 1:
            self._add_error("One time step only.")
            return
        differences = np.diff(time)
        min_difference = np.min(differences)
        max_difference = np.max(differences)
        if min_difference <= 0:
            msg = utils.create_out_of_bounds_msg("time", 0, 24, min_difference)
            self._add_error(msg)
        if max_difference >= 24:
            msg = utils.create_out_of_bounds_msg("time", 0, 24, max_difference)
            self._add_error(msg)


class TestVariableNames(Test):
    name = "Variables"
    description = "Check that file contains required variables."

    def run(self):
        keys_in_file = set(self.nc.variables.keys())
        required_keys = self._get_required_variable_names()
        missing_keys = list(required_keys - keys_in_file)
        for key in missing_keys:
            self._add_error(f"'{key}' is missing.")


class TestModelData(Test):
    name = "Model data"
    description = "Test that model data are valid."
    products = [Product.MODEL]

    def run(self):
        time = np.array(self.nc["time"][:])
        time_unit = datetime.timedelta(hours=1)

        try:
            n_time = len(time)
        except (TypeError, ValueError):
            return
        if n_time < 2:
            return

        duration = self._get_duration()
        should_be_data_until = duration / time_unit

        for key in ("temperature", "pressure", "q"):
            if key not in self.nc.variables:
                continue
            data = self.nc[key][:]
            missing_hours = [
                int(hour)
                for ind, hour in enumerate(time)
                if hour <= should_be_data_until
                and ma.count_masked(data[ind, :]) == data.shape[1]
            ]
            if not missing_hours:
                continue
            noun, verb = ("Hour", "is") if len(missing_hours) == 1 else ("Hours", "are")
            values = utils.format_list(utils.integer_ranges(missing_hours), "and")
            self._add_error(f"{noun} {values} {verb} missing from variable '{key}'.")


# ------------------------------#
# ------ Error / Warning ------ #
# ----------------------------- #


class TestCFConvention(Test):
    name = "CF conventions"
    description = "Test compliance with the CF metadata conventions."

    def run(self):
        from cfchecker import cfchecks  # pylint: disable=import-outside-toplevel

        cf_version = "1.8"
        inst = cfchecks.CFChecker(
            silent=True,
            version=cf_version,
            cfStandardNamesXML=CF_STANDARD_NAMES_XML,
            cfAreaTypesXML=CF_AREA_TYPES_XML,
            cfRegionNamesXML=CF_REGION_NAMES_XML,
        )
        result = inst.checker(str(self.filename))
        for key in result["variables"]:
            for level, error_msg in result["variables"][key].items():
                if not error_msg:
                    continue
                if level in ("FATAL", "ERROR"):
                    severity = ErrorLevel.ERROR
                elif level == "WARN":
                    severity = ErrorLevel.WARNING
                else:
                    continue
                msg = utils.format_msg(error_msg)
                msg = f"Variable '{key}': {msg}"
                self._add_message(msg, severity)


class TestInstrumentPid(Test):
    name = "Instrument PID"
    description = "Test that valid instrument PID exists."
    products = [
        Product.MWR,
        Product.LIDAR,
        Product.RADAR,
        Product.DISDROMETER,
        Product.DOPPLER_LIDAR,
        Product.DOPPLER_LIDAR_WIND,
        Product.WEATHER_STATION,
    ]

    data: dict = {}

    def run(self):
        if self._check_exists():
            try:
                self.data = utils.fetch_pid(self.nc.instrument_pid)
                self._check_serial()
                self._check_model_name()
                self._check_model_identifier()
            except RequestException:
                self._add_info("Failed to fetch instrument PID")

    def _check_exists(self) -> bool:
        key = "instrument_pid"
        try:
            pid = getattr(self.nc, key)
            if pid == "":
                self._add_error("Instrument PID is empty.")
                return False
            if re.fullmatch(utils.PID_FORMAT, pid) is None:
                self._add_error("Instrument PID has invalid format.")
                return False
        except AttributeError:
            self._add_warning("Instrument PID is missing.")
            return False
        return True

    def _get_value(self, kind: str) -> dict | list | None:
        try:
            item = next(item for item in self.data["values"] if item["type"] == kind)
            return json.loads(item["data"]["value"])
        except StopIteration:
            return None

    def _create_message(
        self,
        expected: str | list[str],
        received: str,
        obj: str | None = None,
    ) -> str:
        if isinstance(expected, str):
            expected = [expected]
        expected = utils.format_list([f"'{var}'" for var in expected], "or")
        msg = f"Expected {obj} to be {expected} but received '{received}'"
        return msg

    def _check_serial(self):
        key = "serial_number"
        try:
            received = str(getattr(self.nc, key))
        except AttributeError:
            return
        items = self._get_value("21.T11148/eb3c713572f681e6c4c3")
        if not isinstance(items, list):
            return
        model_name = self._get_value("21.T11148/c1a0ec5ad347427f25d6")["modelName"]
        for item in items:
            if item["alternateIdentifier"]["alternateIdentifierType"] == "SerialNumber":
                expected = item["alternateIdentifier"]["alternateIdentifierValue"]
                if "StreamLine" in model_name:
                    expected = expected.split("-")[-1]
                if received != expected:
                    msg = self._create_message(expected, received, "serial number")
                    self._add_error(msg)
                return
        self._add_warning(
            f"No serial number was defined in instrument PID but found '{received}' in the file."
        )

    def _check_model_name(self):
        key = "source"
        try:
            source = getattr(self.nc, key)
            allowed_values = self.SOURCE_TO_NAME[source]
        except (AttributeError, KeyError):
            return
        model = self._get_value("21.T11148/c1a0ec5ad347427f25d6")
        if model is None:
            return
        received = model["modelName"]
        if received not in allowed_values:
            msg = self._create_message(allowed_values, received, "model name")
            self._add_error(msg)

    def _check_model_identifier(self):
        key = "source"
        try:
            source = getattr(self.nc, key)
            allowed_values = self.SOURCE_TO_IDENTIFIER[source]
        except (AttributeError, KeyError):
            return
        model = self._get_value("21.T11148/c1a0ec5ad347427f25d6")
        if model is None:
            return
        if "modelIdentifier" not in model:
            return
        received = model["modelIdentifier"]["modelIdentifierValue"]
        if received not in allowed_values:
            msg = self._create_message(allowed_values, received, "model identifier")
            self._add_error(msg)

    SOURCE_TO_NAME = {
        "Lufft CHM15k": ["Lufft CHM 15k", "Lufft CHM 15k-x"],
        "Lufft CHM15kx": ["Lufft CHM 15k", "Lufft CHM 15k-x"],
        "TROPOS PollyXT": ["PollyXT"],
        "Vaisala CL31": ["Vaisala CL31"],
        "Vaisala CL51": ["Vaisala CL51"],
        "Vaisala CL61d": ["Vaisala CL61"],
        "Vaisala CT25k": ["Vaisala CT25K"],
        "HALO Photonics StreamLine": [
            "StreamLine",
            "StreamLine Pro",
            "StreamLine XR",
            "StreamLine XR+",
        ],
    }

    SOURCE_TO_IDENTIFIER = {
        "BASTA": ["https://vocabulary.actris.nilu.no/actris_vocab/MeteomodemBASTA"],
        "METEK MIRA-35": [
            "https://vocabulary.actris.nilu.no/actris_vocab/MetekMIRA35",
            "https://vocabulary.actris.nilu.no/actris_vocab/MetekMIRA35S",
            "https://vocabulary.actris.nilu.no/actris_vocab/MetekMIRA35C",
        ],
        "OTT HydroMet Parsivel2": [
            "https://vocabulary.actris.nilu.no/actris_vocab/OTTParsivel2"
        ],
        "RAL Space Copernicus": [
            "https://vocabulary.actris.nilu.no/actris_vocab/RALCopernicus"
        ],
        "RAL Space Galileo": [
            "https://vocabulary.actris.nilu.no/actris_vocab/RALGalileo"
        ],
        "RPG-Radiometer Physics HATPRO": [
            "https://vocabulary.actris.nilu.no/actris_vocab/RPGHATPRO"
        ],
        "RPG-Radiometer Physics RPG-FMCW-35": [
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-35-DP",
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-35-DP-S",
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-35-SP",
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-35-SP-S",
        ],
        "RPG-Radiometer Physics RPG-FMCW-94": [
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-94-DP",
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-94-DP-S",
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-94-SP",
            "https://vocabulary.actris.nilu.no/actris_vocab/RPG-FMCW-94-SP-S",
        ],
        "Thies Clima LNM": ["https://vocabulary.actris.nilu.no/actris_vocab/ThiesLNM"],
    }
